/*
 * PicStamp
 * A simple plugable and breadboardable board
 *
 * Copyright (C) 2012  B.Stultiens
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <stdint.h>
#include <clock.h>
#include <serial.h>

#define WHEEL_TICKS	1000
#define BLINKY_TICKS	333

static const char wheelchars[4] = {'|', '/', '-', '\\'};

/*
 * Interrupt service
 */
void isr(void) __interrupt 1
{
	if(PIE5bits.TMR4IE && PIR5bits.TMR4IF)
		clock_isr();
	if(PIE1bits.RC1IE && PIR1bits.RC1IF)
		serial_recv_isr();
	if(PIE1bits.TX1IE && PIR1bits.TX1IF)
		serial_send_isr();
}

static void delay_100us(void)
{
	__asm
	movlw	177
start_delayloop:
	nop
	nop
	nop
	nop
	nop
	nop
	addlw	255		; w--
	bnz	start_delayloop	; 4/64.0 * 9 * 177 --> ~99.56us
	__endasm;
}

static void delay_ms(uint8_t n)
{
	while(n--) {
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
		delay_100us();
	}
}

int main(void)
{
	static uint8_t wcnt;

	/* Janitorial */
	cli();
	RCON = 0x13;		/* Reset condition */

	/* Setup xtal osc */
	OSCTUNEbits.PLLEN = 1;	/* Set PLL enable to get 64.0MHz */

	/* Delay ~20ms to settle clock */
	delay_ms(20);

	ANSELA = 0x00;		/* 0: Set to digital I/O */
	ANSELB = 0x00;
	ANSELC = 0x00;
	ANSELD = 0x00;
	ANSELE = 0x00;

	SLRCON = 0x00;		/* Standard slewrate on all ports */
	
	PORTA = 0;
	TRISAbits.TRISA0 = 0;	/* LED on PA0 as output */

	wcnt = 0;		/* TIMER_WHEEL character display */

	clock_init();		/* Setup the clock so we may use timing */
	serial_init();		/* Now we can use serial output */

	/* Enable interrupts */
	INTCONbits.PEIE = 1;
	sti();

	/*
	 * Delay before sending serial data to get the host to settle after serial port
	 * init. There may be a change of TX line level that causes framing errors or
	 * break events. The line must be idle for at least one character period.
	 */
	delay_ms(20);

	/* Flush junk and say we are alive */
	serial_puts("\r\n\r\nBooting...\r\n");

	clock_timer_set(TIMER_BLINKY, BLINKY_TICKS);
	clock_timer_set(TIMER_WHEEL, WHEEL_TICKS);

	serial_puts("\r\nPress any key to show uptime and input.\r\n");

	for(;;) {
		if(clock_timer_expired(TIMER_BLINKY)) {
			clock_timer_set(TIMER_BLINKY, BLINKY_TICKS);
			PORTAbits.RA0 ^= 1;
		}

		if(clock_timer_expired(TIMER_WHEEL)) {
			clock_timer_set(TIMER_WHEEL, WHEEL_TICKS);
			serial_putc(wheelchars[wcnt++]);
			serial_putc(0x08);
			wcnt %= sizeof(wheelchars);

		}

		if(serial_have_input()) {
			uint16_t m = clock_msecs;
			serial_putdec_u32(clock_secs);
			serial_putc('.');
			if(m < 100)
				serial_putc('0');
			if(m < 10)
				serial_putc('0');
			serial_putdec_u16(m);
			serial_puts(": You pressed a key: 0x");
			serial_puthex(serial_getc());
			serial_puts("\r\n");
		}

		/* Goto sleep while we have nothing to do */
		OSCCONbits.IDLEN = 1;
		__asm
		sleep
		__endasm;
	}
}
